/*  File: pbwt_impute.c
 *  Author: Richard Durbin (rd@sanger.ac.uk)
 *  Copyright (C) Genome Research Limited, 2013
 *-------------------------------------------------------------------
 * Description:
 * Exported functions:
 * HISTORY:
 * Last edited: Apr 10 15:54 2013 (rd)
 * Created: Thu Apr  4 12:02:56 2013 (rd)
 *-------------------------------------------------------------------
 */

#include "pbwt.h"

/************************* IMPUTATION AND PHASING *************************/

#include <math.h>

void imputeExplore (PBWT *p, int test)
{
  int i, j, k, M = p->M, N = p->N, ff ;
  int ny = 0, nz, c ;
  uchar *y ;
  Update *u = updateCreate (M) ;
  Update *uz = updateCreate (M) ;
  double xbar, ybar, tot, r2 ;
  static double fBound[] = {0.1, 0.2, 0.3, 0.5, 0.7, 1, 2, 3, 4, 5, 7, 10, 20, 50, 100.01} ;
  double f ;
  typedef struct {
    long n00, n01 ;		/* neither neighbour is 1, truth is 0 or 1 */
    long n10a, n10b, n11a, n11b ;/* one neighbour is 1; a if d is lower for the 0 neighbour */
    long n20, n21 ;		/* both neighbours are 1 */
  } TestStat ;
  TestStat *testStat = mycalloc (15, TestStat), *t ;
  typedef long Counts[4] ;
  Array dHist = arrayCreate (1000, Counts) ;
  Counts cSimple, cCond0, cCond1 ;
  int *n0 = myalloc (M, int), *n1 = myalloc (M, int) ;
  uchar *x = myalloc (M, uchar) ;
  static long c0[15][5], c1[15][5] ;

  pbwtBuildReverse (p) ; nz = arrayMax(p->zz) ; memcpy (uz->a, p->za, M*sizeof(int)) ;

  for (i = 0 ; i < 4 ; ++i) cSimple[i] = cCond0[i] = cCond1[i] = 0 ; 

  for (k = 0 ; k < N ; ++k)
    { ny += unpack3 (arrp(p->yz,ny,uchar), M, u->y, &c) ;
      nz -= packCountReverse (arrp(p->zz,nz,uchar), M) ;
      unpack3 (arrp(p->zz,nz,uchar), M, uz->y, 0) ;
      updateBackwardsA (uz, c) ; /* do first so uz->a is other side of y - is this right? */
      if (isCheck)
	{ for (i = 0 ; i < M ; ++i)
	    x[u->a[i]] = u->y[i] ;
	  for (i = 0 ; i < M ; ++i)
	    if (x[uz->a[i]] != uz->y[i]) fprintf (stderr, "at k\n") ;
	}
      if (k > 0.2*N && k < 0.8*N)     /* ignore ends */
	{ f = (M - c) / (double)M ; for (ff = 0 ; f*100 > fBound[ff] ; ++ff) ;
	  t = &testStat[ff] ;
	  memset (n0, 0, M*sizeof(int)) ; memset (n1, 0, M*sizeof(int)) ; 
	  for (i = 1 ; i < M-1 ; ++i)
	    { if (u->y[i-1] && u->y[i+1])
		if (u->y[i]) ++t->n21 ; else ++t->n20 ;
	      else if (!u->y[i-1] && !u->y[i+1])
		if (u->y[i]) ++t->n01 ; else ++t->n00 ;
	      else if (!u->y[i-1] && u->d[i] < u->d[i+1] || !u->y[i+1] && u->d[i+1] < u->d[i])
		if (u->y[i]) ++t->n11a ; else ++t->n10a ;
	      else
		if (u->y[i]) ++t->n11b ; else ++t->n10b ;
	      ++(array(dHist, u->d[i]/100, Counts)[u->y[i-1] + 2*u->y[i]]) ;
	      ++cSimple[u->y[i-1] + 2*u->y[i]] ;
	      if (u->y[i+1])
		++cCond1[u->y[i-1] + 2*u->y[i]] ;
	      else
		++cCond0[u->y[i-1] + 2*u->y[i]] ;
	      n0[u->a[i]] += 2 - (u->y[i-1] + u->y[i+1]) ;
	      n1[u->a[i]] += u->y[i-1] + u->y[i+1] ;
	      n0[uz->a[i]] += 2 - (uz->y[i-1] + uz->y[i+1]) ;
	      n1[uz->a[i]] += uz->y[i-1] + uz->y[i+1] ;
	      x[u->a[i]] = u->y[i] ;
	    }
	  for (i = 0 ; i < M ; ++i)
	    if (n0[i] + n1[i] == 4)	/* it wasn't at the end either forwards or backwards */
	      if (x[i]) ++c1[ff][n1[i]] ; else ++c0[ff][n1[i]] ;
	}
      updateForwardsADU (u, k) ;
    }

  if (test == 1)
    for (j = 0 ; j < 15 ; ++j)
      { t = &testStat[j] ;
	printf ("%-5.1f\t00,01\t%ld\t%ld\t10a,11a\t%ld\t%ld\t10b,11b\t%ld\t%ld\t20,21\t%ld\t%ld",
		fBound[j], t->n00, t->n01, t->n10a, t->n11a, t->n10b, t->n11b, t->n20, t->n21) ;
	tot = t->n00 + t->n01 + t->n10a + t->n11a + t->n10b + t->n11b + t->n20 + t->n21 ;
	if (tot)
	  { xbar = (t->n10b + t->n11b + t->n20 + t->n21)/tot ;
	    ybar = (t->n01 + t->n11a + t->n11b + t->n21)/tot ;
	    r2 = ((t->n21 + t->n11b)/tot - xbar*ybar)/sqrt((xbar - xbar*xbar)*(ybar - ybar*ybar)) ;
	    printf ("\tx,y,r2\t%.4f\t%.4f\t%.4f\n", xbar, ybar, r2) ;
	  }
	else
	  putchar ('\n') ;
      }
  else if (test == 2)
    for (j = 0 ; j < arrayMax(dHist) ; ++j)
      { long *c ; c = arr(dHist, j, Counts) ;
	printf ("%d\t%ld\t%ld\t%ld\t%ld", j, c[0], c[1], c[2], c[3]) ;
	if (c[0] + c[2]) printf ("\t%.3f", c[0]/(double)(c[0]+c[2])) ; else printf ("\t0") ;
	if (c[1] + c[3]) printf ("\t%.3f", c[3]/(double)(c[1]+c[3])) ; else printf ("\t0") ;
	putchar ('\n') ;
      }
  else if (test == 3)
    { printf ("%.3f %.3f\t", cSimple[0]/(double)(cSimple[0]+cSimple[2]), cSimple[3]/(double)(cSimple[1]+cSimple[3])) ;
      printf ("%.3f %.3f\t", cCond0[0]/(double)(cCond0[0]+cCond0[2]), cCond0[3]/(double)(cCond0[1]+cCond0[3])) ;
      printf ("%.3f %.3f\n", cCond1[0]/(double)(cCond1[0]+cCond1[2]), cCond1[3]/(double)(cCond1[1]+cCond1[3])) ;
    }
  else if (test == 4)
    for (j = 0 ; j < 15 ; ++j)
      { printf ("%-5.1f", fBound[j]) ;
	tot = 0 ; xbar = 0 ; r2 = 0 ;
        for (i = 0 ; i < 5 ; ++i)
	  { long sum = c0[j][i] + c1[j][i] ;
	    printf ("\t%ld ", sum) ;
	    if (sum) printf (" %.3f", c1[j][i]/(double)sum) ; else printf (" 00000") ;
	    xbar += c1[j][i] ;
	    if (i == 3 || i == 4) r2 += c1[j][i] ; if (i ==2) r2 += 0.5*c1[j][i] ;
	    tot += sum ;
	  }
	ybar = c0[j][4] + c1[j][4] + c0[j][3] + c1[j][3] + 0.5*(c0[j][2] + c1[j][2]) ;
	if (tot)
	  { xbar /= tot ;
	    ybar /= tot ;
	    r2 = (r2/tot - xbar*ybar)/sqrt((xbar - xbar*xbar)*(ybar - ybar*ybar)) ;
	    printf ("\tx,y,r2\t%.4f\t%.4f\t%.4f\n", xbar, ybar, r2) ;
	  }
	else
	  putchar ('\n') ;
      }

  updateDestroy (u) ;
}

/****************** phasing ****************/

static void phaseCompare (PBWT *p, PBWT *q, int *nSwitch, int *nHet, int *nSwitch0)
{
  int i, k ;
  int M = p->M, N = p->N ;
  Update *up = updateCreate (M), *uq = updateCreate (M) ;
  int nyp = 0, nyq = 0 ;
  int *xp = myalloc (M, int), *xq = myalloc (M, int) ;
  int *isFirst = myalloc (M, int), *isFlipped = myalloc (M, int) ;
  int kMin = N/10 ;

  if (p->M != q->M || p->N != q->N) die ("size incompatibility in phaseCompare") ;
  if (p->M %2) die ("phaseCompare requires that M %d is even", M) ;

  memset (isFirst, 1, M*sizeof(int)) ; /* I know this is not 1 - anything non-zero will do */
  *nSwitch = 0 ; *nHet = 0 ; *nSwitch0 = 0 ;
  for (k = 0 ; k < N ; k++)
    { nyp += unpack3 (arrp(p->yz,nyp,uchar), M, up->y, 0) ;
      nyq += unpack3 (arrp(q->yz,nyq,uchar), M, uq->y, 0) ;
      for (i = 0 ; i < M ; ++i)
	{ xp[up->a[i]] = up->y[i] ;
	  xq[uq->a[i]] = uq->y[i] ;
	}
      for (i = 0 ; i < M ; i += 2)
	{ if (xp[i] + xp[i+1] == 1)
	    { ++*nHet ;
	      if (isFirst[i])
		{ isFirst[i] = 0 ;
		  isFlipped[i] = (xp[i] == xq[i+1]) ? 1 : 0 ; 
		}
	      else if (xp[i] != xq[i+isFlipped[i]]) 
		{ ++*nSwitch ;
		  if (k > kMin) ++*nSwitch0 ;
		  isFlipped[i] = 1 - isFlipped[i] ;
		}
	    }
	  if (isCheck && (xp[i]+xp[i+1] != xq[i]+xq[i+1])) die ("phaseCompare mismatch") ;
	}
      updateForwardsA (up) ;
      updateForwardsA (uq) ;
    }

  updateDestroy (up) ; updateDestroy (uq) ;
  free (xp) ; free (xq) ; free (isFirst) ; free (isFlipped) ;
}

void phaseTest (PBWT *p, int nCycle)
{
  int i, k, c ;
  int M = p->M, N = p->N ;
  PBWT *q = pbwtCreate (M), *r = pbwtCreate (M) ;
  Update *up = updateCreate (M), *uq = updateCreate (M), *ur = updateCreate (M) ;
  int nyp = 0, nyq, nyr, n2, n2Old ;
  int *x = myalloc (M, int), *aq = myalloc (M, int), *ar = myalloc (M, int), *pair = myalloc (M, int) ;
  int nSwitch, nHet, nSwitch0 ;
  uchar *yz = myalloc(M, uchar) ;
  static int score[3] = { -1, 1, 0 } ;
  
  if (p->M %2) die ("phaseTest requires that M = %d is even", M) ;

  q->yz = arrayCreate (arrayMax(p->yz), uchar) ; q->N = N ;

  for (k = 0 ; k < N ; k++)
    { nyp += unpack3 (arrp(p->yz,nyp,uchar), M, up->y, 0) ;
      for (i = 0 ; i < M ; ++i)	/* in p order */
	{ x[up->a[i]] = up->y[i] ;
	  aq[uq->a[i]] = i ;	/* inverse of uq->a */
	}
      n2 = 0 ;
      for (i = 0 ; i < M ; i += 2) /* in x order */
	if (x[i] + x[i+1] == 1)
	  { uq->y[aq[i]] = 2 ; uq->y[aq[i+1]] = 2 ; ++n2 ;  /* use 2 as unresolved symbol */
	    pair[aq[i]] = aq[i+1] ; pair[aq[i+1]] = aq[i] ; 
	  }
	else
	  { uq->y[aq[i]] = x[i] ; uq->y[aq[i+1]] = x[i+1] ; 
	  }
      n2Old = M+1 ;
      /* next iteratively resolve 2s by neighbour matches where possible */
      while (n2 && n2 < n2Old)	/* stop if no longer reducing unresolved symbols */
	{ n2Old = n2 ; n2 = 0 ;
	  for (i = 0 ; i < M ; ++i) /* in q order */
	    if (uq->y[i] == 2)
	      { int s = 0 ;
		if (i > 0) s += score[uq->y[i-1]] ;
		if (i < M-1) s += score[uq->y[i+1]] ;
		if (pair[i] > 0) s -= score[uq->y[pair[i]-1]] ;
		if (pair[i] < M-1) s -= score[uq->y[pair[i]+1]] ;
		if (s > 0) { uq->y[i] = 1 ; uq->y[pair[i]] = 0 ; }
		else if (s < 0) { uq->y[i] = 0 ; uq->y[pair[i]] = 1 ; }
		else ++n2 ;
	      }
	}
      if (n2)			/* some unresolved values - set first to 0, second to 1 */
	for (i = 0 ; i < M ; ++i) /* in q order */
	  if (uq->y[i] == 2)
	    { uq->y[i] = 0 ; uq->y[pair[i]] = 1 ; }
      
      updateForwardsA (up) ;
      updateForwardsA (uq) ;
      nyq = pack3 (uq->y, M, yz) ;
      for (i = 0 ; i < nyq ; ++i) array(q->yz,arrayMax(q->yz),uchar) = yz[i] ;
    }
  
  phaseCompare (p, q, &nSwitch, &nHet, &nSwitch0) ;
  printf ("%d hets\n", nHet) ;
  printf ("After forward phase pass: %.1f switches per sample, %.3f per het, %.1f in last 0.9\n", 
	  nSwitch/(double)M, nSwitch/(double)nHet, nSwitch0/(double)M) ;

  /* now go backwards, running back through q, uq and making r, ur in the reverse direction */

  r->yz = arrayCreate (arrayMax(q->yz), uchar) ; r->N = N ;
  memcpy (ur->a, uq->a, M*sizeof(int)) ;
  nyq = arrayMax(q->yz) ;
  for (k = N ; k-- ; )
    { nyq -= packCountReverse (arrp(q->yz, nyq, uchar), M) ;
      unpack3 (arrp(q->yz, nyq, uchar), M, uq->y, &c) ;

      updateBackwardsA (uq, c) ; /* this needs to come before the next section? */
      for (i = 0 ; i < M ; ++i)
	{ x[uq->a[i]] = uq->y[i] ;
	  ar[ur->a[i]] = i ;	/* inverse of ur->a */
	}

      n2 = 0 ;
      for (i = 0 ; i < M ; i += 2) /* in x order */
	if (x[i] + x[i+1] == 1)
	  { ur->y[ar[i]] = 2 ; ur->y[ar[i+1]] = 2 ; ++n2 ;  /* use 2 as unresolved symbol */
	    pair[ar[i]] = ar[i+1] ; pair[ar[i+1]] = ar[i] ; 
	  }
	else
	  { ur->y[ar[i]] = x[i] ; ur->y[ar[i+1]] = x[i+1] ; }

      n2Old = M+1 ;
      /* next iteratively resolve 2s by neighbour matches where possible */
      while (n2 && n2 < n2Old)	/* stop if no longer reducing unresolved symbols */
	{ n2Old = n2 ; n2 = 0 ;
	  for (i = 0 ; i < M ; ++i)
	    if (ur->y[i] == 2)
	      { int s = 0 ;
		if (i > 0) s += score[ur->y[i-1]] ;
		if (i < M-1) s += score[ur->y[i+1]] ;
		if (pair[i] > 0) s -= score[ur->y[pair[i]-1]] ;
		if (pair[i] < M-1) s -= score[ur->y[pair[i]+1]] ;
		if (s > 0) { ur->y[i] = 1 ; ur->y[pair[i]] = 0 ; }
		else if (s < 0) { ur->y[i] = 0 ; ur->y[pair[i]] = 1 ; }
		else ++n2 ;
	      }
	}
      if (n2)			/* some unresolved values - set first to 0, second to 1 */
	for (i = 0 ; i < M ; ++i) /* in q order */
	  if (uq->y[i] == 2)
	    { uq->y[i] = 0 ; uq->y[pair[i]] = 1 ; }

      updateForwardsA (ur) ;
      nyr = pack3 (ur->y, M, yz) ;
      for (i = 0 ; i < nyr ; ++i) array(q->yz,arrayMax(q->yz),uchar) = yz[i] ;
    }


  updateDestroy (up) ; updateDestroy (uq) ; pbwtDestroy (q) ;
  free (x) ; free (aq) ; free (pair) ; free (yz) ;
}

/******************* end of file *******************/
